-module(node).
-include("../include/tsp.hrl").
-compile(export_all).

start(MasterNode, State) ->
	DFSL = State#state.starting_level,
	Best = State#state.best,
	BestLength = Best#tour.length,
	Problem = State#state.problem,
	Master = {?MASTER, MasterNode},
	io:format("~p started. Reporting to ~p.~n", [self(), Master]),
	Master ! {fetch, self()},
	loop(BestLength, {Master, DFSL, Problem}).

loop(BL, {Master, _DFSL, _Problem} = State) ->
	receive
		die ->
			io:format("Received die.~n"),
			dead;
		Request ->
			NewBest = handle(Request, BL, State),
			Master ! {fetch, self()},
			loop(NewBest, State)
	end.

% Received a new job
handle({job, Tour}, BL, {_Master, _DFSL, Problem} = State) ->
	%io:format("~p received a new job~n", [self()]),
	Children = tour:get_children(Tour, Problem),
	NewBest = solve(Children, BL, State),
	%io:format("~p is done solving.~n", [self()]),
	NewBest.

solve([], BL, _State) ->
	% No more children to process
	BL;
solve([#tour{pending = []} = Child|Children], BL, {Master, _DFSL, _Problem} = State) ->
	% It's a full tour. Offer it to the server, if it's better than the local best
	NewBest = case Child#tour.length < BL of
		true ->
			io:format("~p: Offer full solution: L = ~p~n", [self(), Child#tour.length]),
			Master ! {{offer, Child}, self()},
			receive
				{new_best_length, Update} -> Update
			end;
		false ->
			BL
	end,
	%io:format("~p: Received updated best length: ~p~n", [self(), NewBest]),
	solve(Children, NewBest, State);
solve([#tour{fixed = Fixed} = Child|Children], BL, {_Master, DFSL, _Problem} = State)
	when length(Fixed) >= DFSL ->
	% Incomplete tour, that has a length over DFSL
	%io:format("~p: Solve depth first.~n", [self()]),
	NewBest = dfsolve(Child, BL, State),
	solve(Children, NewBest, State);
solve([#tour{lower_bound = LB} = Child|Children], BL, {Master, _DFSL, _Problem} = State) ->
	% Incomplete tour, with a length < DFSL
	% If it has a usable lower bound, offer it to the server
	NewBL = case LB < BL of
		true ->
			% The child might be a solution
			%io:format("~p: Offer partial solution.~n", [self()]),
			Master ! {{offer, Child}, self()},
			receive
				{new_best_length, NewBest} -> NewBest
			end;
		false ->
			% Child has a worse LB than best length, ignore it
			BL
	end,
	solve(Children, NewBL, State).


dfsolve(#tour{pending = [], length = L} = Tour, BL, {Master, _DFSL, _Problem})
	when L < BL ->
	% Full tour, with a length better than best length
	io:format("~p: Offer full solution: L = ~p~n", [self(), L]),
	Master ! {{offer, Tour}, self()},
	receive
		{new_best_length, NewBest} -> NewBest
	end;
dfsolve(#tour{lower_bound = LB} = Tour, BL, {_Master, _DFSL, Problem} = State) ->
	case LB > BL of
		true ->
			% This child cannot give a better solution
			BL;
		false ->
			%io:format("~p: dfsolve: BL: ~p, LB: ~p...~n", [self(), BL, LB]),
			Children = tour:get_children(Tour, Problem),
			solve(Children, BL, State)
	end.
	
