-module(tour).
-compile(export_all).
-include("../include/tsp.hrl").

create_tour(Cities, Problem) ->
	Pending = Problem#problem.cities -- Cities,
	update_lower_bound(update_length(#tour{fixed = Cities, pending = Pending}, Problem), Problem).

create_tour_nocalc(Cities, Problem) ->
	Pending = Problem#problem.cities -- Cities,
	#tour{fixed = Cities, pending = Pending}.

get_children(#tour{fixed = F, pending = P}, Problem) ->
	get_children(F, P, Problem, []).

get_children(_Fixed, [], _Problem, Children) ->
	Children;
get_children(Fixed, [City|Pending], Problem, Children) ->
	Child = create_tour([City|Fixed], Problem),
	get_children(Fixed, Pending, Problem, [Child|Children]).

update_lower_bound(#tour{length = L, pending = []} = Tour, _Problem) ->
	% If is_full_tour(Tour) ->
	Tour#tour{lower_bound = L};
update_lower_bound(#tour{pending = P, fixed = F, length = L} = Tour, #problem{distances = D} = Problem) ->
	{FirstId, _} = hd(F),
	{LastId, _} = lists:last(F),
	BoundFirst = lists:min([dict:fetch(order_pair({FirstId, ThisId}), D) || {ThisId, _} <- P]),
	BoundLast = lists:min([dict:fetch(order_pair({LastId,ThisId}), D) || {ThisId, _} <- P]),
	LowerBound = (BoundFirst + BoundLast) / 2.0,
	update_lower_bound(Tour#tour{lower_bound = L + LowerBound}, Problem, FirstId, LastId, P).

update_lower_bound(Tour, _Problem, _FirstId, _LastId, []) ->
	% No more cities to process
	Tour;
update_lower_bound(#tour{pending = P, lower_bound = LB} = Tour, #problem{distances = D} = Problem, FirstId, LastId, [City|Cities]) ->
	{CurrId, _} = City,
	Bound1 = dict:fetch(order_pair({CurrId, FirstId}), D),
	Bound2 = dict:fetch(order_pair({CurrId, LastId}), D),
	Bounds = [Bound1, Bound2 | [dict:fetch(order_pair({CurrId, OtherId}), D) || {OtherId, _} <- (P -- [City])]],
	% Find the two smallest distances if we add this city
	MinBound1 = lists:min(Bounds),
	MinBound2 = lists:min(Bounds -- [MinBound1]),
	LowerBound = (MinBound1 + MinBound2) / 2.0,
	update_lower_bound(Tour#tour{lower_bound = LB + LowerBound}, Problem, FirstId, LastId, Cities).


update_length(#tour{fixed = [_]} = Tour, _Problem) ->
	Tour#tour{length = 0};
update_length(#tour{fixed = Cities, pending = []} = Tour, #problem{distances = D} = Problem) ->
	% Full tour, so get distance from last to first city
	{FirstId, _} = hd(Cities),
	{LastId, _} = lists:last(Cities),
	Length = dict:fetch(order_pair({LastId, FirstId}), D),
	% Get the rest of the distances
	update_length(Tour#tour{length = Length}, Problem, Cities);
update_length(#tour{fixed = Cities} = Tour, Problem) ->
	update_length(Tour#tour{length = 0}, Problem, Cities).

update_length(Tour, _Problem, [_]) ->
	% No more cities to process
	Tour;
update_length(#tour{length = Length} = Tour, #problem{distances = D} = Problem, [City1, City2|Cities]) ->
	% Order the pair of ids
	{From,To} = order_pair({element(1, City1), element(1, City2)}),
	% Fetch the distance between City1 and City2
	Length1 = Length + dict:fetch({From, To}, D),
	update_length(Tour#tour{length=Length1}, Problem, [City2|Cities]).

create_initial_tours(#state{problem = P} = State) ->
	[First|Cities] = P#problem.cities,
	create_initial_tours(First, Cities, State, []).

create_initial_tours(_First, [_Second], _State, Tours) ->
	Tours;
create_initial_tours(First, [Second|Cities], State, Tours) ->
	NewTours = create_initial_tours2(First, Second, Cities, State, []),
	create_initial_tours(First, Cities, State, Tours ++ NewTours).

create_initial_tours2(_One, _Two, [], _State, Tours) ->
	Tours;
create_initial_tours2(One, Two, [Three|Cities], #state{problem = P} = State, Tours) ->
	Tour = tour:create_tour([One,Two,Three], P),
	Best = State#state.best,
	NewTours = case Tour#tour.lower_bound < Best#tour.lower_bound of
		true -> [Tour|Tours];
		false -> Tours
	end,
	create_initial_tours2(One, Two, Cities, State, NewTours).


order_pair({A,B}) when A > B -> {B,A};
order_pair(Pair) -> Pair.
