-module(tsp).
-include("../include/tsp.hrl").
-compile(export_all).

ready() ->
	register(?MASTER, spawn(fun() -> wait(waiting) end)),
	io:format("Ready...~n").

set(Procs, Nodes, DepthFirstStartingLevel, Filename) ->
	P = problem:load_from_file(Filename),
	Best = greedy:solve(P),
	InitState = #state{procs = Procs, nodes = Nodes,
		starting_level = DepthFirstStartingLevel, problem = P,
		alive = Procs * length(Nodes), best = Best},
	Initial = tour:create_initial_tours(InitState),
	Q = lists:foldl(fun pq:push/2, pq:empty(), Initial),
	State = InitState#state{queue = Q},
	spawn_procs(Procs, Nodes, State),
	tsp ! {set_state, State},
	io:format("Initial best tour: ~p~n", [Best#tour.length]),
	io:format("We're set...~n").

go() ->
	tsp ! start,
	io:format("And, we're going: ~p~n", [time()]).

spawn_procs(NumProcs, Nodes, State) ->
	NodesAndNums = [{Node, NumProcs, State} || Node <- Nodes],
	lists:foreach(fun spawn_procs/1, NodesAndNums).
spawn_procs({_Node, 0, _State}) ->
	true;
spawn_procs({Node, NumProcs, State}) ->
	spawn_link(Node, node, start, [node(), State]),
	spawn_procs({Node, NumProcs - 1, State}).

wait(CurrState) ->
	receive
		{set_state, State} -> wait(State);
		start -> loop(CurrState)
	end.


loop(#state{alive = 0} = State) ->
	Best = State#state.best,
	io:format("Found an optimal solution with length ~p.~n", [Best#tour.length]),
	io:format("Tour: ~p.~n", [Best#tour.fixed]),
	io:format("Time: ~p.~n", [time()]);
loop(State) ->
	receive
		{Request, From} ->
			{Response, NewState} = handle(Request, State),
			From ! Response,
			loop(NewState)
	end.


handle(status, State) ->
	io:format("Best current: ~p~n", [State#state.best]),
	{ok, State};
% A node requests a new job
handle(fetch, #state{best = B, queue = Q} = State) ->
	%io:format("Received a fetch request~n"),
	case pq:take_usable (B#tour.length, Q) of
		{Tour, NewQ} ->
			NewState = State#state{queue = NewQ},
			{{job, Tour}, NewState};
		empty ->
			Alive = State#state.alive,
			{die, State#state{alive = Alive - 1}}
	end;

%handle({offer, _}, State) ->
%	io:format("Received offer~n"),
%	{State#state.best, State};

% This is the case where we are offered a full tour
handle({offer, #tour{pending = []} = Tour}, #state{best = B} = State) ->
	%io:format("Received a full tour~n"),
	NewBest = case Tour#tour.length < B#tour.length of
		true ->
			io:format("New best tour: ~p~n", [Tour#tour.length]),
			Tour;
		false ->
			B
	end,
	{{new_best_length, NewBest#tour.length}, State#state{best = NewBest}};
% We are offered a partial tour.
handle({offer, PartTour}, #state{best = B, queue = Q} = State) ->
	% The node has already checked if lb is below its own best,
	% but the current best might be different
	io:format("Received a partial tour~n"),
	NewQ = case PartTour#tour.lower_bound < B#tour.length of
		true ->	pq:push(PartTour, Q);
		false -> Q
	end,
	{{new_best_length, B#tour.length}, State#state{queue = NewQ}}.



